"""
File        : mkdocs_table_of_figures/structure/type_figure/__init__.py (type_figure.py)
Description : This file contains the FigureType and FigureTypesCollection classes.
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

from typing import Dict, List, Optional, Any

from ...translations import Translations
from ..collection import Collection

class FigureType:
    """
    Represents a type of figure with a pattern used to find it and a template used to replace it.
    """
    # Constants
    TEMPLATE_DEFAULT = \
'''<figure id="{{ figure.id }}" class="{{ figure.classes }}" markdown="block">
{{ figure.match_content }}
<figcaption markdown="span">{{ figure.caption }}</figcaption>
</figure>'''
    TRANSLATIONS_DEFAULT = None
    METADATA_DEFAULT = {}

    # Class attributes
    translations: Optional[Translations] = TRANSLATIONS_DEFAULT

    # Attributes
    name: str
    pattern: str
    template: str
    metadata: Dict[str, Any]

    # Properties
    @property
    def label(self) -> str:
        output = self.name
        
        # If translation exist use it instead of default value
        if self.translations and self.translations[f'figure_types.{self.name}']:
            output = self.translations[f'figure_types.{self.name}']

        return output

    # Constructors
    def __init__(self, name: str, pattern: str, template: str = TEMPLATE_DEFAULT, metadata: Optional[Dict[str, Any]] = None):
        self.name = name
        self.pattern = rf'{pattern}'
        self.template = rf'{template}'
        self.metadata = metadata if metadata is not None else dict(self.METADATA_DEFAULT) # Ensure creation of a copy of default dict
    
    @classmethod
    def init_from_config(cls, config: Dict[str, Any]) -> "FigureType":
        # NOTE: config.get('template', cls.TEMPLATE_DEFAULT) don't work it return 'None'
        return cls(config['name'], config['pattern'], config.get('template') or cls.TEMPLATE_DEFAULT, config.get('metadata'))

    # Method
    def __str__(self) -> str:
        return f"<FigureType name={self.name} pattern=r'{self.pattern}'>"

class FigureTypes(Collection[FigureType]):
    """
    This is a Collection of FigureType with additional properties 'names'.
    """
    # Constants
    ITEMS_STR_PREVIEW_SEPARATOR = '|'

    # Attributes
    # ...
    
    # Properties
    @property
    def names(self) -> List[str]:
        return [item.name for item in self]
    
    # Constructors
    # ...

    # Class Methods
    @classmethod
    def init_from_config(cls, config: List[Any]) -> "FigureTypes":
        items = []

        for entry in config:
            items.append(FigureType.init_from_config(entry))
        
        return cls(items)

    # Methods
    # ...



"""
File        : mkdocs_table_of_figures/config.py
Description : This is the config class containing the plugin config schema 
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

import os
import yaml
from typing import Dict

from mkdocs.config.base import Config, BaseConfigOption, ValidationError
from mkdocs.config import config_options

# Import to resolve path of builtin figure types configs files (.yml)
from .structure.figure_type import builtin_configs

class FigureTypeConfig(Config):
    name = config_options.Type(str)
    pattern = config_options.Type(str)
    template = config_options.Optional(config_options.Type(str))
    metadata = config_options.Optional(config_options.Type(dict))

class FigureTypeOption(BaseConfigOption):
    SUBCONFIG = config_options.SubConfig(FigureTypeConfig)
    BUILTIN_TYPES_DIRECTORY = os.path.dirname(builtin_configs.__file__)

    _builtin_types: Dict[str, FigureTypeConfig]

    @property
    def builtin_types(self) -> Dict[str, FigureTypeConfig]:
        if not self._builtin_types:
            self._load_builtin_types()
        
        return self._builtin_types
    
    def __init__(self):
        self._builtin_types = {}

    def _load_builtin_types(self) -> None:
        self._builtin_types = {}
        
        for filename in os.listdir(self.BUILTIN_TYPES_DIRECTORY):
            if os.path.isfile(os.path.join(self.BUILTIN_TYPES_DIRECTORY, filename)) and os.path.splitext(filename)[1] in ['.yml', '.yaml']:
                name = os.path.splitext(filename)[0]
                path = os.path.join(self.BUILTIN_TYPES_DIRECTORY, filename)
                with open(path, 'r', encoding='utf8') as file:
                    try:
                        config = yaml.safe_load(file)
                        if isinstance(config, dict):
                            if 'name' not in config.keys():
                                config['name'] = name
                            self._builtin_types[name] = config
                    except yaml.YAMLError as e:
                        raise ValidationError(f"Error loading built-in type '{filename}': {e}")

    def run_validation(self, value):
        if isinstance(value, str):
            if value in self.builtin_types:
                return self.SUBCONFIG.validate(self.builtin_types[value])
            else:
                raise ValidationError(f"Unknown built-in figure type '{value}'. Available: {list(self.builtin_types.keys())}")
        elif isinstance(value, dict):
            return self.SUBCONFIG.validate(value)
        else:
            raise ValidationError(f"Expected a string or a dict for FigureTypeConfig, got: {value!r}")

class TableOfFiguresConfig(Config):
    DIRECTIVE_IDENTIFIER_DEFAULT = 'table-of-figures'
    FIGURE_TYPES_DEFAULT = ['image', 'table', 'codeblock']

    CUSTOM_TEMPLATES_PATH_DEFAULT = None
    TRANSLATIONS_OVERRIDES_PATH_DEFAULT = None

    CAPTION_IDENTIFIER_DESCRIPTION_SEPARATOR_DEFAULT = ' — '
    SECTIONS_BREADCRUMB_SEPARATOR_DEFAULT = ' > '
    
    directive_identifier = config_options.Type(str, default=DIRECTIVE_IDENTIFIER_DEFAULT)
    figure_types = config_options.ListOfItems(FigureTypeOption(), default=FIGURE_TYPES_DEFAULT)

    custom_templates_path = config_options.Optional(config_options.Type(str, default=CUSTOM_TEMPLATES_PATH_DEFAULT))
    translations_overrides_path = config_options.Optional(config_options.Type(str, default=TRANSLATIONS_OVERRIDES_PATH_DEFAULT))

    caption_identifier_description_separator = config_options.Type(str, default=CAPTION_IDENTIFIER_DESCRIPTION_SEPARATOR_DEFAULT)
    sections_breadcrumb_separator = config_options.Type(str, default=SECTIONS_BREADCRUMB_SEPARATOR_DEFAULT)



"""
File        : mkdocs_table_of_figures/structure/figure.py
Description : This file contains the Figure and FiguresCollection classes,
              along with a special exception thrown when
              a figure match is tempered before being processed.
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

from re import Match
from typing import Dict, Optional, Any
from jinja2 import Template

from mkdocs.structure.pages import Page

from ..translations import Translations
from .figure_type import FigureType
from .collection import Collection

class Figure:
    """
    This represents a figure with its index, page, type, match, replacement, caption, etc.
    """
    # Constants
    CAPTION_IDENTIFIER_PREFIX_DEFAULT = 'Figure '
    CAPTION_IDENTIFIER_SUFFIX_DEFAULT = ''

    TRANSLATIONS_DEFAULT = None

    ID_PREFIX_DEFAULT = '__figure_'
    CLASSES_PREFIX_DEFAULT = 'figure figure-'
    CAPTION_IDENTIFIER_DESCRIPTION_SEPARATOR_DEFAULT = ' - '
    SECTIONS_BREADCRUMB_SEPARATOR_DEFAULT = ' / '

    # Class attributes
    translations: Optional[Translations] = TRANSLATIONS_DEFAULT

    id_prefix: str = ID_PREFIX_DEFAULT
    classes_prefix: str = CLASSES_PREFIX_DEFAULT
    caption_identifier_description_separator: str = CAPTION_IDENTIFIER_DESCRIPTION_SEPARATOR_DEFAULT
    sections_breadcrumb_separator: str = SECTIONS_BREADCRUMB_SEPARATOR_DEFAULT

    # Attributes
    index: int
    figure_type: FigureType
    caption_description: str
    page: Page
    match: Optional[Match]

    # Properties
    @property
    def id(self) -> str:
        return f'{self.id_prefix}{self.index}'
    
    @property
    def classes(self) -> str:
        return f'{self.classes_prefix}{self.figure_type_name}'
    
    @property
    def caption_identifier_prefix(self) -> str:
        output = self.CAPTION_IDENTIFIER_PREFIX_DEFAULT
        
        # If translation exist use it instead of default value
        if self.translations and self.translations['caption.index.prefix']:
            output = self.translations['caption.index.prefix']

        return output
    
    @property
    def caption_identifier_suffix(self) -> str:
        output = self.CAPTION_IDENTIFIER_SUFFIX_DEFAULT
        
        # If translation exist use it instead of default value
        if self.translations and self.translations['caption.index.suffix']:
            output = self.translations['caption.index.suffix']

        return output
    
    @property
    def caption_identifier(self) -> str:
        return f'{self.caption_identifier_prefix}{self.index}{self.caption_identifier_suffix}'
    
    @property
    def caption(self) -> str:
        output = f'{self.caption_identifier}'

        # If an empty caption was provided only show identifier (without separator)
        if self.caption_description:
            output += f'{self.caption_identifier_description_separator}{self.caption_description}'
        
        return output
    
    @property
    def figure_type_name(self) -> str:
        return self.figure_type.name
    
    @property
    def figure_type_label(self) -> str:
        return self.figure_type.label
    
    @property
    def figure_type_metadata(self) -> Dict[str, Any]:
        return self.figure_type.metadata
    
    @property
    def page_uri(self) -> str:
        return f'{self.page.file.src_uri}'
    
    @property
    def page_header(self) -> str:
        output = ''

        for line in self.page.file.content_string.splitlines():
            if line.startswith("# "):
                output = line[2:].strip()
                break
        
        return output
    
    @property
    def sections_breadcrumb(self) -> str:
        output = self.page.title or self.page_header

        section = self.page.parent
        while section:
            output = f'{section.title}{self.sections_breadcrumb_separator}{output}'
            section = section.parent

        return output
    
    @property
    def uri(self) -> str:
        return f'{self.page_uri}#{self.id}'
    
    @property
    def match_content(self) -> str:
        return self.match.group(0)
    
    @property
    def match_start(self) -> int:
        return self.match.start()
    
    @property
    def match_end(self) -> int:
        return self.match.end()
    
    @property
    def match_groups(self) -> dict:
        return self.match.groupdict()
    
    @property
    def replacement(self) -> str:
        context = {'figure': self}
        template = Template(self.figure_type.template)

        return template.render(context)
    
    # Constructors
    def __init__(self, index: int, figure_type: FigureType, caption_description: str, page: Page, match: Optional[Match] = None):
        self.index = index
        self.figure_type = figure_type
        self.caption_description = caption_description
        self.page = page
        self.match = match
    
    # Methods
    def __str__(self) -> str:
        f'<Figure index={self.index} type={self.figure_type_name} caption={self.caption_description}>'

class FiguresCollection(Collection[Figure]):
    """
    This is a Collection of Figure.
    """
    def page_figures(self, page: Page) -> 'FiguresCollection':
        figures = [figure for figure in self if figure.page == page]
        figures.sort(key=lambda f: f.index) # Ensure figure are sorted by position in the page
        return FiguresCollection(figures)

class FigureChangedException(Exception):
    """
    Exception raised when a figure match has been tampered with before processing.
    """
    # Nothing change from base Exception
    pass



"""
File        : mkdocs_table_of_figures/structure/collection.py
Description : This file contains the base class Collection
              used by FiguresCollection, FigureTypesCollection and DirectivesCollection.
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

from typing import List, Iterator, Optional, TypeVar, Generic

T = TypeVar('T')

class Collection(Generic[T]):
    """
    A generic base class representing a list-like collection of objects of the same type.
    """
    # Constants
    ITEMS_DEFAULT = []

    ITEMS_STR_PREVIEW_NUMBER = 3
    ITEMS_STR_PREVIEW_SEPARATOR = ', '
    ITEMS_STR_PREVIEW_MORE = '...'

    # Attributes
    items: List[T]

    # Properties
    @property
    def size(self) -> int:
        return len(self.items)
    
    @property
    def _items_str_preview(self) -> str:
        output = self.ITEMS_STR_PREVIEW_SEPARATOR.join(repr(item) for item in self.items[:3])

        # Only output the first few items, adding suspension if there are more
        if self.size > self.ITEMS_STR_PREVIEW_NUMBER:
            output += f'{self.ITEMS_STR_PREVIEW_SEPARATOR}{self.ITEMS_STR_PREVIEW_MORE}'
        
        return output

    # Constructors
    def __init__(self, items: Optional[List[T]] = None):
        self.items = items if items is not None else list(self.ITEMS_DEFAULT) # Ensure creation of a copy of default list

    # Methods
    def __iter__(self) -> Iterator[T]:
        return iter(self.items)
    
    def __getitem__(self, index: int) -> T:
        return self.items[index]
    
    def __str__(self) -> str:
        return f'<{self.__class__.__name__} size={self.size} items=[{self._items_str_preview}]>'

    def append(self, item: T):
        self.items.append(item)



"""
File        : mkdocs_table_of_figures/constants.py
Description : Theses are some constants from the package (same as setup.cfg)
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

from . import __version__

PACKAGE_NAME = 'mkdocs-table-of-figures'
VERSION = __version__

# Parse entrypoint identifier
MKDOCS_ENTRYPOINT = 'table-of-figures'



"""
File        : mkdocs_table_of_figures/builder.py
Description : This is the builder class that will be used to build markdown content in documentation using Jinja2 templates
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

import os
from typing import Optional

from jinja2 import Environment, Template, FileSystemLoader

class Builder:
    SCRIPT_DIR_PATH = os.path.dirname(os.path.abspath(__file__))
    CUSTOM_TEMPLATES_PATH_DEFAULT = None
    TEMPLATES_PATH_FALLBACK = os.path.join(SCRIPT_DIR_PATH, 'templates/')
    OUTPUT_TEMPLATE_FILE = 'output.md.j2'

    templates_path: str
    loader: FileSystemLoader
    env: Environment

    @property
    def output_template(self) -> Template:
        return self.env.get_template(self.OUTPUT_TEMPLATE_FILE)

    def __init__(self, templates_path: str):
        self.templates_path = templates_path
        self.loader = FileSystemLoader(self.templates_path)
        self.env = Environment(loader=self.loader)

    @classmethod
    def init_from_config(cls, custom_templates_path: Optional[str] = CUSTOM_TEMPLATES_PATH_DEFAULT) -> 'Builder':
        template_path = custom_templates_path or cls.TEMPLATES_PATH_FALLBACK

        return Builder(template_path)


    def build(self, context: dict) -> str:
        return self.output_template.render(context)



"""
File        : mkdocs_table_of_figures/translations/__init__.py (translations.py)
Description : This file contains the translations handler.
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

import os
import json

from typing import Dict, Union, Optional

class TranslationsEntries:
    """
    A tree-like structure holding translation entries, supporting dotted key access
    """
    # Constants
    ENTRIES_KEY_SEPARATOR = '.'

    # Attributes
    entries: Dict[str, Union[str, 'TranslationsEntries']]

    # Properties
    # ...

    # Constructors
    def __init__(self, entries: Dict[str, Union[str, dict]]):
        self.entries = {}

        for key, value in entries.items():
            if isinstance(value, dict):
                value = TranslationsEntries(value)
            self.entries[key] = value

    # Methods
    def __getitem__(self, key) -> Optional[Union[str, 'TranslationsEntries']]:
        """
        Supports dot-separated key lookup for nested entries.
        """

        def _recursive_getitem(item: 'TranslationsEntries', key) -> Optional[Union[str, 'TranslationsEntries']]:
            value = item.entries.get(key)

            if self.ENTRIES_KEY_SEPARATOR in key:
                current_key, rest_keys = key.split(self.ENTRIES_KEY_SEPARATOR, 1)
                value = item.entries.get(current_key)
                
                if isinstance(value, TranslationsEntries) and rest_keys:
                    value = _recursive_getitem(value, rest_keys)

            return value

        return _recursive_getitem(self, key)


class Translations(TranslationsEntries):
    """
    Translation handler that loads language files, supports fallbacks,
    and allows entry override at load time.
    """
    # Constants
    TRANSLATIONS_DIRECTORY = os.path.dirname(__file__)
    TRANSLATIONS_FILE_FORMAT = 'json'
    FALLBACK_LANG = 'en'
    USE_FALLBACK_DEFAULT = True

    # Attributes
    lang: str

    # Properties
    # ...

    # Constructors
    def __init__(self, lang: str, entries: Dict[str, Union[str, dict]]):
        super().__init__(entries)
        self.lang = lang

    # NOTE: These 2 constructors are a little bit confusing. Might be good to change them.
    @classmethod
    def init_from_json(cls, lang: str, use_fallback: bool = USE_FALLBACK_DEFAULT, *overrides: Dict[str, Union[str, dict]]) -> 'Translations':
        fallback_dict = cls._load_json_file(cls._get_lang_file(cls.FALLBACK_LANG)) if use_fallback else {}
        lang_dict = cls._load_json_file(cls._get_lang_file(lang))

        merged_dict = cls._deep_merge(fallback_dict, lang_dict)
        for override in overrides:
            merged_dict = cls._deep_merge(merged_dict, override)

        return cls(lang, merged_dict)
    
    @classmethod
    def init_from_config(cls, lang: str, overrides_path: str) -> 'Translations':
        overrides_dict = cls._load_json_file(overrides_path) if overrides_path is not None else {}

        return cls.init_from_json(lang, cls.USE_FALLBACK_DEFAULT, overrides_dict)
    
    # Class methods
    @classmethod
    def _get_lang_file(cls, lang: str) -> str:
        return os.path.join(cls.TRANSLATIONS_DIRECTORY, f'{lang}.{cls.TRANSLATIONS_FILE_FORMAT}')

    @classmethod
    def _load_json_file(cls, filepath: str) -> Dict[str, Union[str, dict]]:
        data = {}

        if os.path.isfile(filepath):
            with open(filepath, "r", encoding="utf-8") as f:
                data = json.load(f)

            if not isinstance(data, dict):
                raise ValueError(f"{filepath} must contain a top-level JSON object")
        
        return data
    
    @classmethod
    def _deep_merge(cls, base: Dict[str, Union[str, dict]], override: Dict[str, Union[str, dict]]) -> Dict[str, Union[str, dict]]:
        result = base.copy()

        for key, value in override.items():
            if key in result and isinstance(result[key], dict) and isinstance(value, dict):
                result[key] = cls._deep_merge(result[key], value)
            else:
                result[key] = value

        return result



"""
File        : mkdocs_table_of_figures/structure/directive.py
Description : This file contains the Directive and DirectivesCollection classes,
              and a custom decorator to define class-level property.
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

import re
from typing import List

from mkdocs.structure.pages import Page as MkDocsPage

from ..constants import MKDOCS_ENTRYPOINT
from .collection import Collection

class classproperty(property):
    """
    Custom decorator to define a class-level property.
    """
    def __get__(self, obj, cls):
        return self.fget(cls)

class Directive:
    """
    A directive block detected in the Markdown content.
    """
    PATTERN_OPENING = r'<!--\s*?'
    PATTERN_CLOSING = r'\s*?-->'

    identifier: str = MKDOCS_ENTRYPOINT

    @classproperty
    def pattern(cls) -> str:
        return rf'{cls.PATTERN_OPENING}(?P<identifier>{cls.identifier})(?:\s*?(?P<options>\[.*?\]))?{cls.PATTERN_CLOSING}'

    options: 'DirectiveOptions'
    context: 'DirectiveContext'

    def __init__(self, options: 'DirectiveOptions', context: 'DirectiveContext'):
        self.options = options
        self.context = context
    
    @classmethod
    def init_from_match(cls, page: MkDocsPage, match: re.Match) -> 'Directive':
        options = DirectiveOptions()
        context = DirectiveContext(page, match)

        return cls(options, context)

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} options={self.options} context={self.context}>'

class DirectiveOptions:
    """
    Class for options users may provide to customize rendering.
    """
    # TODO Provides options for users to customize the rendering
    def __init__(self):
        pass

class DirectiveContext:
    """
    Contextual information about where the directive appears in the page.
    """
    # Constants
    # ...

    # Attributes
    page: MkDocsPage
    match: re.Match

    # Properties
    @property
    def page_content(self) -> str:
        return self.page.file.content_string
    
    @property
    def page_content_before(self) -> str:
        return self.page_content[:self.start]
    
    @property
    def page_lines_before(self) -> List[str]:
        return self.page_content_before.splitlines()
    
    @property
    def page_content_after(self) -> str:
        return self.page_content[self.end:]
    
    @property
    def page_lines_after(self) -> List[str]:
        return self.page_content_after.splitlines()
    
    @property
    def span(self) -> tuple[int, int]:
        return self.match.span()

    @property
    def start(self) -> int:
        return self.match.start()
    
    @property
    def end(self) -> int:
        return self.match.end()

    @property
    def start_line(self) -> int:
        # Count number of line breaks before the end index
        return 1 + self.page_content[:self.start].count('\n')
    
    @property
    def start_column(self) -> int:
        # Column are counted from the last \n before the match
        return self.start - self.page_content[:self.start].rfind('\n')
    
    @property
    def end_line(self) -> int:
        # Count number of line breaks before the end index
        return 1 + self.page_content[:self.end].count('\n')

    @property
    def end_column(self) -> int:
        # Column is offset from the last newline before the end index
        return self.end - self.page_content[:self.end].rfind('\n')
    
    @property
    def match_index(self) -> int:
        return len(list(re.finditer(Directive.pattern, self.page_content_before)))
    
    @property
    def current_heading_level(self) -> int:
        value = 0

        for line in reversed(self.page_lines_before):
            stripped = line.strip()
            if stripped.startswith("#"):
                match_heading = re.match(r'^(#+)\s+', stripped)
                if match_heading:
                    value = len(match_heading.group(1))
        
        return value
    
    # Constructors
    def __init__(self, page: MkDocsPage, match: re.Match):
        self.page = page
        self.match = match

    # Methods

class DirectivesCollection(Collection[Directive]):
    """
    A collection of Directive objects with helpers to extract and filter them.
    """
    # Constants, Attributes, Properties, Constructors
    # ...

    # Methods
    def appends_from_markdown(self, page: MkDocsPage, markdown: str):
        for match in re.finditer(Directive.pattern, markdown):
            self.append(Directive.init_from_match(page, match))

    def filter_by_page(self, target_page: MkDocsPage) -> 'DirectivesCollection':
        filtered = DirectivesCollection()
        for directive in self:
            if directive.context.page == target_page:
                filtered.append(directive)
        return filtered



"""
File        : table_of_figures/plugin.py
Description : This is the mkdocs plugin that interface with mkdocs
Author      : Thibaud Briard - BRT, <thibaud.briard@outlook.com>
"""

import os
import re

from mkdocs.plugins import BasePlugin
from mkdocs.plugins import get_plugin_logger

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.pages import Page as MkDocsPage
from mkdocs.structure.files import Files as MkDocsFiles, File as MkDocsFile
from mkdocs.structure.nav import Section as MkDocsSection, Link as MkDocsLink

from .constants import MKDOCS_ENTRYPOINT

from .structure.directive import Directive, DirectivesCollection
from .structure.figure_type import FigureType, FigureTypes
from .structure.figure import Figure, FiguresCollection, FigureChangedException
from .translations import Translations

from .builder import Builder
from .config import TableOfFiguresConfig as PluginConfig

class TableOfFiguresPlugin(BasePlugin[PluginConfig]):
    PLUGIN_ENTRYPOINT = MKDOCS_ENTRYPOINT
    STYLES_DIRECTORY = 'assets/stylesheets/figures/'

    def __init__(self):
        self.logger = get_plugin_logger(__name__)

        self.config_dir = None
        self.custom_templates_dir = None
        self.translations_overrides_file = None

        self.builder = None

        self.translations = None
        
        self.figure_types = None
        self.figure_extra_css = None

        self.directives = None
        self.figures = None

    def on_config(self, config: MkDocsConfig) -> MkDocsConfig | None:
        self.logger.debug(self.config)

        # Retrieve user customs files and directories
        self.config_dir = os.path.dirname(config.config_file_path)
        if self.config.custom_templates_path:
            self.custom_templates_dir = os.path.normpath(os.path.join(self.config_dir, self.config.custom_templates_path))
        if self.config.translations_overrides_path:
            self.translations_overrides_file = os.path.normpath(os.path.join(self.config_dir, self.config.translations_overrides_path))

        # Set template builder with pottential user templates
        self.builder = Builder.init_from_config(self.custom_templates_dir)
        # Set translations according to theme language and pottential user overrides
        self.translations = Translations.init_from_config(config.theme.get('language') or config.theme.get('local') or 'none', self.config.translations_overrides_path)

        # Set Calsses attributes according to mkdocs config
        Directive.identifier = self.config.directive_identifier
        FigureType.translations = self.translations
        Figure.translations = self.translations
        Figure.caption_identifier_description_separator = self.config.caption_identifier_description_separator
        Figure.sections_breadcrumb_separator = self.config.sections_breadcrumb_separator

        # Retrieve figure types from figure type subconfig
        self.figure_types = FigureTypes.init_from_config(self.config.figure_types)
        self.logger.debug(f'Figures types [{self.figure_types.size}]: {self.figure_types.names}')

        # Initialize figure_extra_css list
        self.figure_extra_css = []

        # Collect existing CSS files based on enabled figure types
        for figure_type in self.figure_types:
            figure_css = os.path.join(self.STYLES_DIRECTORY, f'{figure_type.name}.css')
            if os.path.isfile(os.path.join(os.path.dirname(__file__), figure_css)):
                self.figure_extra_css.append(figure_css)

        # Prepend these existing CSS files to config['extra_css']
        # This ensure plugin CSS rules can be overriden by user CSS rules
        config['extra_css'] = self.figure_extra_css + config.get('extra_css', [])

        # Set
        self.directives = DirectivesCollection()
        self.figures = FiguresCollection()

        return config
    
    def on_nav(self, nav, config: MkDocsConfig, files: MkDocsFiles):
        def process_nav_item(item: MkDocsSection|MkDocsPage|MkDocsLink):
            if isinstance(item, MkDocsSection) and hasattr(item, 'children') and item.children:
                # It's a Section
                self.logger.debug(f'Processing childs of nav item (section): {item}')
                for child in item.children:
                    process_nav_item(child)
            elif isinstance(item, MkDocsPage) and hasattr(item, 'file') and item.file:
                # It's a Page
                self.logger.debug(f'Processing nav item (page): {item}')
                process_page(item)
            else:
                self.logger.debug(f'Skipping nav item: {item}')
        
        def process_page(page: MkDocsPage):
            def check_for_overlapsing_matches(existing_matches, target_match):
                """
                Check if the target match overlaps with existing matches.
                It prevent overlapsing figures that have similar pattern.
                """
                return any(existing_match.start() < target_match.end() and target_match.start() < existing_match.end() for _, existing_match, _ in existing_matches)
            
            matches = []

            # Add matches for each type of figure and order them by appearance
            for figure_type in self.figure_types:
                for match in re.finditer(figure_type.pattern, page.file.content_string):
                    if not check_for_overlapsing_matches(matches, match):
                        self.logger.debug(f'Found matching figure "{figure_type.name}" at [{match.start()}:{match.end()}]')
                        matches.append((match.start(), match, figure_type))
                    else:
                        self.logger.debug(f'Found an overlapsing matching figure "{figure_type.name}" at [{match.start()}:{match.end()}], It will be ignored')
            matches.sort(key=lambda x: x[0])

            # Loop trough each match and create a Figure entry (also saving match for futur replacement)
            for _, match, figure_type in matches:
                # Create new Figure entry to collection 
                new_figure = Figure(self.figures.size + 1, figure_type, match.group('caption'), page)
                self.figures.append(new_figure)
                self.logger.debug(f'New figure added to collection: {new_figure.page_uri}[{match.start()}:{match.end()}] - {new_figure.figure_type_name} "{new_figure.caption_description}"')

        for item in nav.items:
            process_nav_item(item)

    def on_files(self, files: MkDocsFiles, config: MkDocsConfig):
        # Add plugin styles to MkDocs files collection
        for css_file in self.figure_extra_css:
            file = MkDocsFile(path=css_file, src_dir=os.path.dirname(__file__), dest_dir=config['site_dir'], use_directory_urls=config['use_directory_urls'])
            # Mark file as generated by this plugin
            file.generated_by = self.PLUGIN_ENTRYPOINT
            files.append(file)

        return files
    
    def on_page_markdown(self, markdown: str, page: MkDocsPage, config: MkDocsConfig, files: MkDocsFiles) -> str | None:
        file_uri = page.file.src_uri
        offset = 0

        # Retrieve all figures matches in the page content and order them by occurrence
        matches = []
        for figure_type in self.figure_types:
            for match in re.finditer(figure_type.pattern, page.file.content_string):
                if not any(existing_match.start() < match.end() and match.start() < existing_match.end() for existing_match in matches):
                    matches.append(match)
        matches.sort(key=lambda x: x.start())

        checkpoint_markdown = markdown
        try:
            # Adding figure to page
            for index, figure in enumerate(self.figures.page_figures(page)):
                # Add match to figure object
                figure.match = matches[index]

                # Check that figure content in markdown content haven't been tempered with
                if figure.match_content != markdown[figure.match_start + offset:figure.match_end + offset]:
                    raise FigureChangedException(f'The figure ({figure.index}) in {page} as been tempered with. modification to this page be ignored')

                self.logger.debug(f'Replacing figure [{figure.index}]:')
                [self.logger.debug(f'| - {line}') for line in figure.match_content.splitlines()]
                [self.logger.debug(f'| + {line}') for line in figure.replacement.splitlines()]

                markdown = markdown[:figure.match_start + offset] + figure.replacement + markdown[figure.match_end + offset:]
                offset += len(figure.replacement) - len(figure.match_content)
        except FigureChangedException as error:
            # Reset modification in case of error
            markdown = checkpoint_markdown
            offset = 0
            self.logger.warning(error)

        self.logger.debug(f'Adding table of figures for each directive found')
        self.directives.appends_from_markdown(page, markdown)

        for directive in self.directives.filter_by_page(page):
            replacement = self.builder.build({'figures': self.figures, 'directive': directive, 'translations': self.translations, 'config': config})
            
            self.logger.debug(f'Placing table of figures:')
            self.logger.debug(f'| - {markdown[directive.context.start:directive.context.end]}')
            [self.logger.debug(f'| + {line}') for line in replacement.splitlines()]

            markdown = markdown[:directive.context.start] + replacement + markdown[directive.context.end:]
            offset += len(replacement) - (directive.context.end - directive.context.start)

        return markdown
